#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import logging
import simplejson as json
import os
import shutil
import subprocess
import time
from threading import Thread
from threading import Event
from threading import RLock

_logger = logging.getLogger("sanji.sdk.model_initiator")


class ModelInitiator(object):
    """
    "   Deal with some model initialization works like DB
    "   and Condifuration files creating.
    "   [backup_inteval]: backup db every hours(s), minus means no backup.
    """
    def __init__(
            self, model_name, model_path, db_type="json", backup_interval=720):
        self.model_name = model_name
        self.model_path = model_path
        self.db = None
        self.data_folder_path = self.model_path + "/data"
        self.factory_json_db_path = self.model_path + "/data/" + \
            self.model_name + ".json.factory"
        self.backup_json_db_path = self.model_path + "/data/" + \
            self.model_name + ".json.backup"
        self.json_db_path = self.model_path + "/data/" + \
            self.model_name + ".json"
        self.db_type = db_type
        self.db_status = None
        self.backup_interval = backup_interval * 3600  # hour
        self.db_mutex = RLock()
        self.db_manager()

        self._backup_thread = Thread(target=self.thread_backup_db)
        self._backup_thread.daemon = True
        self._backup_thread_event = Event()
        if self.backup_interval > 0:
            self.start_backup()

    def db_manager(self):
        """
        " Do series of DB operations.
        """
        rc_create = self.create_db()    # for first create
        try:
            self.load_db()  # load existing/factory
        except Exception as e:
            _logger.debug("*** %s" % str(e))
            try:
                self.recover_db(self.backup_json_db_path)
            except Exception:
                pass
        else:
            if rc_create is True:
                self.db_status = "factory"
            else:
                self.db_status = "existing"
            return True

        try:
            self.load_db()  # load backup
        except Exception as b:
            _logger.debug("*** %s" % str(b))
            self.recover_db(self.factory_json_db_path)
            self.load_db()  # load factory
            self.db_status = "factory"
        else:
            self.db_status = "backup"
        finally:
            return True

    def create_db(self):
        """
        "   Create a db file for model if there is no db.
        "   User need to prepare thier own xxx.json.factory.
        """
        if self.db_type != "json":
            raise RuntimeError("db_type only supports json now")

        if os.path.exists(self.json_db_path):
            return False

        if os.path.exists(self.factory_json_db_path):
            with self.db_mutex:
                shutil.copy2(
                    self.factory_json_db_path, self.json_db_path)
            return True

        _logger.debug(
            "*** NO such file: %s" % self.factory_json_db_path)
        raise RuntimeError("No *.json.factory file")

    def recover_db(self, src_file):
        """
        " Recover DB from xxxxx.backup.json or xxxxx.json.factory to xxxxx.json
        " [src_file]: copy from src_file to xxxxx.json
        """
        with self.db_mutex:
            try:
                shutil.copy2(src_file, self.json_db_path)
            except IOError as e:
                _logger.debug("*** NO: %s file." % src_file)
                raise e

    def backup_db(self):
        """
        " Generate a xxxxx.backup.json.
        """
        with self.db_mutex:
            if os.path.exists(self.json_db_path):
                try:
                    shutil.copy2(self.json_db_path, self.backup_json_db_path)
                except (IOError, OSError):
                    _logger.debug("*** No file to copy.")

    def load_db(self):
        """
        " Load json db as a dictionary.
        """
        try:
            with open(self.json_db_path) as fp:
                self.db = json.load(fp)
        except Exception as e:
            _logger.debug("*** Open JSON DB error.")
            raise e

    def save_db(self):
        """
        " Save json db to file system.
        """
        with self.db_mutex:
            if not isinstance(self.db, dict) and not isinstance(self.db, list):
                return False
            try:
                with open(self.json_db_path, "w") as fp:
                    json.dump(self.db, fp, indent=4)
            except Exception as e:
                # disk full or something.
                _logger.debug("*** Write JSON DB to file error.")
                raise e

            else:
                self.sync()
                return True

    def start_backup(self):
        if self._backup_thread.is_alive():
            raise RuntimeError("Stop previous backup thread first.")

        self._backup_thread = Thread(target=self.thread_backup_db)
        self._backup_thread.daemon = True
        self._backup_thread.start()
        return True

    def stop_backup(self, timeout=None):
        if self._backup_thread.is_alive():
            self._backup_thread_event.set()
            if timeout:
                self._backup_thread.join(timeout)
            else:
                self._backup_thread.join()
            return True
        return False

    def thread_backup_db(self):
        single_sleep_time = 2
        sleep_count = self.backup_interval
        while not self._backup_thread_event.is_set():
            if sleep_count >= self.backup_interval:
                self.backup_db()
                sleep_count = 0
            else:
                time.sleep(single_sleep_time)
                sleep_count += single_sleep_time

    def sync(self):
        """
        " Call Linux 'sync' command to write data from RAM to flash.
        """
        cmd = "sync"
        subprocess.call(cmd, shell=True)



import simplejson as json
import copy
import os
import random
from random import randint

try:
    from urllib import parse
except ImportError:
    import urlparse as parse

random.seed(os.urandom(10))


def parse_querystring(querystring):
    """
    Return parsed querystring in dict
    """
    if querystring is None or len(querystring) == 0:
        return {}

    qs_dict = parse.parse_qs(querystring, keep_blank_values=True)
    for key in qs_dict:
        if len(qs_dict[key]) != 1:
            continue
        qs_dict[key] = qs_dict[key][0]
        if qs_dict[key] == '':
            qs_dict[key] = True

    return dict((key, qs_dict[key]) for key in qs_dict if len(key) != 0)


def trim_resource(resource):
    """
    trim_resource
    """
    return resource.strip(" \t\n\r/")


class MessageType(object):
    """
    Message Type Enum

        Check is response message or not
        - [v] id
        - [v] code
        - [v] method
        - [v] resource
        - [v] sign
        - [x] tunnel

        Check is request message or not
        - [v] id
        - [x] code
        - [v] method
        - [v] resource
        - [x] sign
        - [x] tunnel

        Check is direct message or not
        - [v] id
        - [x] code
        - [v] method
        - [v] resource
        - [x] sign
        - [v] tunnel

        Check is event message or not
        - [x] id
        - [v] code
        - [v] method
        - [v] resource
        - [x] sign
        - [x] tunnel

        Check is hook message or not
        - [v] id
        - [x] code
        - [v] method
        - [v] resource
        - [v] sign
        - [x] tunnel
    """

    UNKNOWN = 0
    RESPONSE = 1
    REQUEST = 2
    DIRECT = 3
    EVENT = 4
    HOOK = 5

    FIELDS = {
        RESPONSE: {
            "must": ["id", "code", "method", "resource", "sign"],
            "prohibit": ["tunnel"]
        },
        REQUEST: {

            "must": ["id", "method", "resource"],
            "prohibit": ["code", "sign", "tunnel"]
        },
        DIRECT: {
            "must": ["id", "method", "resource", "tunnel"],
            "prohibit": ["code", "sign"]
        },
        EVENT: {
            "must": ["code", "method", "resource"],
            "prohibit": ["id", "sign", "tunnel"]
        },
        HOOK: {
            "must": ["id", "method", "resource", "sign"],
            "prohibit": ["code", "tunnel"]
        }
    }


class Message(object):
    """
    Message
    """
    def __init__(self, message, generate_id=False):
        if isinstance(message, str) or isinstance(message, unicode):
            try:
                message = json.loads(message)
            except Exception:
                raise ValueError("Invaild Message." +
                                 "Must be a vaild JSON String")

        if not isinstance(message, dict):
            raise TypeError("Message must be JSON string or Dict")

        # put all prop into object
        for (prop, value) in message.iteritems():
            setattr(self, prop, value)

        if generate_id is True:
            self.generate_id()

        # put message type
        self._type = Message.get_message_type(self.__dict__)

    def generate_id(self):
        setattr(self, "id", randint(0, 655350))
        return self.id

    def type(self):
        return self._type

    def to_json(self, pretty=True):
        """
        to_json will call to_dict then dumps into json format
        """
        data_dict = self.to_dict()
        if pretty:
            return json.dumps(
                data_dict, sort_keys=True, indent=2)
        return json.dumps(data_dict, sort_keys=True)

    def to_dict(self):
        """
        to_dict will clean all protected and private properties
        """
        return dict(
            (k, self.__dict__[k]) for k in self.__dict__ if k.find("_") != 0)

    def match(self, route):
        """
        Match input route and return new Message instance
        with parsed content
        """
        _resource = trim_resource(self.resource)
        self.method = self.method.lower()
        resource_match = route.resource_regex.search(_resource)
        if resource_match is None:
            return None

        # build params and querystring
        params = resource_match.groupdict()
        querystring = params.pop("querystring", "")
        setattr(self, "param", params)
        setattr(self, "query", parse_querystring(querystring))

        return copy.deepcopy(self)

    def to_response(self, sign, code=200, data=None):
        """
        transform message to response message
        Notice: this method will return a deepcopy
        """
        msg = copy.deepcopy(self)
        msg.data = data

        setattr(msg, 'code', code)
        for _ in ["query", "param", "tunnel"]:
            if not hasattr(msg, _):
                continue
            delattr(msg, _)

        if hasattr(msg, 'sign') and isinstance(msg.sign, list):
            msg.sign.append(sign)
        else:
            msg.sign = [sign]

        msg._type = Message.get_message_type(msg.__dict__)

        return msg

    def to_event(self):
        """
        get rid of id, sign, tunnel and update message type
        Notice: this method will return a deepcopy
        """
        msg = copy.deepcopy(self)
        for _ in ["id", "sign", "tunnel", "query", "param"]:
            if not hasattr(msg, _):
                continue
            delattr(msg, _)

        msg._type = Message.get_message_type(msg.__dict__)

        return msg

    @staticmethod
    def get_message_type(message):
        """
        Return message's type
        """
        for msg_type in MessageType.FIELDS:
            if Message.is_type(msg_type, message):
                return msg_type

        return MessageType.UNKNOWN

    @staticmethod
    def is_type(msg_type, msg):
        """
        Return message's type is or not
        """
        for prop in MessageType.FIELDS[msg_type]["must"]:
            if msg.get(prop, False) is False:
                return False
        for prop in MessageType.FIELDS[msg_type]["prohibit"]:
            if msg.get(prop, False) is not False:
                return False

        return True



"""
Publish message module
"""

import logging
from time import sleep

from sanji.message import Message
from sanji.session import Status
from sanji.session import TimeoutError
from sanji.session import StatusError


_logger = logging.getLogger("sanji.sdk.publish")


class Object(object):
    pass


class Publish(object):

    """
    Publish class
    """

    def __init__(self, connection, session):
        self._conn = connection
        self._session = session

        self.direct = Object()
        self.event = Object()
        for method in ["get", "post", "put", "delete"]:
            self.__setattr__(method, self.create_crud_func(method))
            self.direct.__setattr__(method,
                                    self.create_crud_func(method, "DIRECT"))

        for method in ["get", "post", "put", "delete"]:
            self.event.__setattr__(method,
                                   self.create_event_func(method))

    def _wait_resolved(self, session):
        session["is_resolved"].wait()
        if session["status"] == Status.RESPONSE_TIMEOUT:
            raise TimeoutError(session)
        elif session["status"] == Status.RESOLVED:
            return session["resolve_message"]
        raise StatusError(session)

    def _wait_published(self, session, no_response=False):
        session["is_published"].wait()
        if session["status"] == Status.SEND_TIMEOUT:
            raise TimeoutError(session)
        elif session["status"] == Status.SENT:
            # if no_resonse is required, we resolve it as empty response data
            if no_response is True:
                self._session.resolve(session["message"].id)
            return session
        raise StatusError(session)

    def _create_message(self, headers=None, data=None, generate_id=True):
        payload = headers
        if isinstance(data, Message):
            return data
        else:
            if data is not None:
                payload["data"] = data

        return Message(payload, generate_id=generate_id)

    def create_event_func(self, method):
        def _crud(resource,
                  data=None,
                  code=200,
                  timeout=60,
                  topic="/controller",
                  qos=2):

            message = self._create_message(
                headers={"resource": resource, "method": method, "code": code},
                data=data,
                generate_id=False)

            with self._session.session_lock:
                mid = self._conn.publish(topic=topic, qos=qos,
                                         payload=message.to_dict())
                session = self._session.create(message, mid=mid, age=timeout)
                session["status"] = Status.SENDING
            return self._wait_published(session, no_response=True)
        return _crud

    def create_crud_func(self, method, request_type="CRUD"):
        """
        create_crud_func
        """
        def _crud(resource,
                  data=None,
                  block=True,
                  timeout=60,
                  topic="/controller",
                  tunnel=None,
                  qos=2):
            """
            _crud

            block
                True: wait until response arrival
                False: wait until message is already published to local broker
            """
            headers = {
                "resource": resource,
                "method": method
            }

            # DIRECT message needs put tunnel in headers for controller
            if request_type == "DIRECT":
                if tunnel is not None:
                    headers["tunnel"] = tunnel
                elif self._conn.tunnels["view"][0] is not None:
                    headers["tunnel"] = self._conn.tunnels["view"][0]
                elif self._conn.tunnels["model"][0] is not None:
                    headers["tunnel"] = self._conn.tunnels["model"][0]
                else:
                    headers["tunnel"] = self._conn.tunnels["internel"][0]

            message = self._create_message(headers, data)
            with self._session.session_lock:
                mid = self._conn.publish(topic=topic,
                                         qos=qos,
                                         payload=message.to_dict())
                session = self._session.create(message, mid=mid, age=timeout)
                session["status"] = Status.SENDING

            # blocking, until we get response or published
            if block is False:
                return self._wait_published(session)
            return self._wait_resolved(session)
        return _crud

    def create_response(self, message, sign):
        """
        return function for response
        """
        def _response(code=200, data=None):
            """
            _response
            """
            resp_msg = message.to_response(code=code, data=data, sign=sign)

            with self._session.session_lock:
                mid = self._conn.publish(topic="/controller",
                                         qos=2, payload=resp_msg.to_dict())
                session = self._session.create(resp_msg, mid=mid, age=10)
            logging.debug("sending response as mid: %s" % mid)
            return self._wait_published(session, no_response=True)
        return _response


def Retry(target=None, args=[], kwargs={},
          options={"retry": True, "interval": 1}):
    """
    options
        retry
            True, infinity retries
            False, no retries
            Number, retries times
        interval
            time period for retry
        return
            None if no success
            Message if success
    """
    retry = options["retry"]
    interval = options["interval"]

    while True:
        try:
            resp = target(*args, **kwargs)
            # status error
            if resp.code == 200:
                return resp

            _logger.debug("Request got response status: %s"
                          % (resp.code,) + " retry: %s" % (retry,))
        except TimeoutError:
            _logger.debug("Request message is timeout")
            _logger.debug(args)
            _logger.debug(kwargs)

        # register unsuccessful goes here
        # infinity retry
        if retry is True:
            sleep(interval)
            continue

        # no retry
        if retry is False:
            return None

        # retrying
        try:
            retry = retry - 1
            if retry <= 0:
                return None
        except TypeError as e:
            raise e
        sleep(interval)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
from __future__ import print_function

import os
import sys
import uuid
import logging
import simplejson as json
import paho.mqtt.client as mqtt
from time import sleep


try:
    sys.path.append(os.path.dirname(os.path.realpath(__file__)) + '/../../')
    from sanji.connection.connection import Connection
except ImportError as e:
    print(e)
    print("Please check the python PATH for import test module.")
    exit(1)


_logger = logging.getLogger("sanji.sdk.connection.mqtt")


class Mqtt(Connection):

    """
    Mqtt
    """
    def __init__(
        self,
        broker_host=os.getenv('BROKER_PORT_1883_TCP_ADDR', "localhost"),
        broker_port=os.getenv('BROKER_PORT_1883_TCP_PORT', 1883),
        broker_keepalive=60
    ):
        # proerties
        self.tunnels = {
            "internel": (uuid.uuid4().hex, None),
            "model": (None, None),
            "view": (None, None)
        }
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.broker_keepalive = broker_keepalive
        self.client = mqtt.Client()
        self.connect_delay = 3

        # methods
        self.subscribe = self.client.subscribe
        self.unsubscribe = self.client.unsubscribe
        self.message_callback_add = self.client.message_callback_add
        self.message_callback_remove = self.client.message_callback_remove
        self.client.on_log = self.on_log

    def on_log(self, mosq, obj, level, string):
        pass

    def connect(self):
        """
        connect
        """
        _logger.debug("Start connecting to broker")
        while True:
            try:
                self.client.connect(self.broker_host, self.broker_port,
                                    self.broker_keepalive)
                break
            except Exception:
                _logger.debug(
                    "Connect failed. wait %s sec" % self.connect_delay)
                sleep(self.connect_delay)

        self.client.loop_forever()

    def disconnect(self):
        """
        disconnect
        """
        _logger.debug("Disconnect to broker")
        self.client.loop_stop()

    def set_tunnel(self, tunnel_type, tunnel, callback=None):
        """
        set_tunnel(self, tunnel_type, tunnel, callback=None):
        """
        orig_tunnel = self.tunnels.get(tunnel_type, (None, None))[0]
        if orig_tunnel is not None:
            _logger.debug("Unsubscribe: %s", (orig_tunnel,))
            self.client.unsubscribe(str(orig_tunnel))

        self.tunnels[tunnel_type] = (tunnel, callback)

        if callback is not None:
            self.message_callback_add(tunnel, callback)

        self.client.subscribe(str(tunnel))
        _logger.debug("Subscribe: %s", (tunnel,))

    def set_tunnels(self, tunnels):
        """
        set_tunnels(self, tunnels):
        """
        for tunnel_type, (tunnel, callback) in tunnels.iteritems():
            if tunnel is None:
                continue
            self.set_tunnel(tunnel_type, tunnel, callback)

    def set_on_connect(self, func):
        """
        set_on_connect
        """
        self.client.on_connect = func

    def set_on_message(self, func):
        """
        set_on_message
        """
        self.client.on_message = func

    def set_on_publish(self, func):
        """
        set_on_publish
        """
        self.client.on_publish = func

    def publish(self, topic="/controller", qos=2, payload=None):
        """
        publish(self, topic, payload=None, qos=0, retain=False)
        Returns a tuple (result, mid), where result is MQTT_ERR_SUCCESS to
        indicate success or MQTT_ERR_NO_CONN if the client is not currently
        connected.  mid is the message ID for the publish request. The mid
        value can be used to track the publish request by checking against the
        mid argument in the on_publish() callback if it is defined.
        """
        result = self.client.publish(topic,
                                     payload=json.dumps(payload),
                                     qos=qos)
        if result[0] == mqtt.MQTT_ERR_NO_CONN:
            raise RuntimeError("No connection")
        return result[1]



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
import simplejson as json
import logging

_logger = logging.getLogger("sanji.sdk.bundle")


class Bundle(object):

    def __init__(self,
                 bundle_dir=os.getenv("BUNDLE_DIR", os.getcwd()),
                 sdk_dir=os.path.dirname(os.path.realpath(__file__)),
                 load=True):
        self.sdk_dir = sdk_dir
        self.bundle_dir = bundle_dir
        self.profile = None
        if load is True:
            self.load_json()

    def load_json(self):
        json_path = os.path.join(self.bundle_dir, "bundle.json")
        try:
            with open(json_path) as f:
                self.profile = json.load(f)
        except Exception as e:
            raise BundleConfigError("Can't load file: %s, error: %s",
                                    (json_path, str(e)))

        _logger.debug("Model: %s config has been loaded." %
                      (self.profile["name"],))


class BundleConfigError(Exception):
    pass



"""
Connection Interface Module
"""


class Connection(object):
    """
    Connection interface
    """
    def connect(self):
        """
        connect
        """
        raise NotImplementedError("Should have implemented this")

    def disconnect(self):
        """
        disconnect
        """
        raise NotImplementedError("Should have implemented this")

    def set_tunnel(self, type, tunnel, callback=None):
        """
        set tunnel
        """
        raise NotImplementedError("Should have implemented this")

    def set_tunnels(self, tunnels):
        """
        set tunnels
        """
        raise NotImplementedError("Should have implemented this")

    def set_on_connect(self, func):
        """
        set_on_connect
        """
        raise NotImplementedError("Should have implemented this")

    def set_on_message(self, func):
        """
        set_on_message
        """
        raise NotImplementedError("Should have implemented this")

    def set_on_publish(self, func):
        """
        set_on_publish
        """
        raise NotImplementedError("Should have implemented this")

    def publish(self, *args, **kwargs):
        """
        publish
        """
        raise NotImplementedError("Should have implemented this")



"""
This is router class
"""

import re
from sanji.message import trim_resource


def compile_resource(resource):
    """
    Return compiled regex for resource matching
    """
    return re.compile("^" + trim_resource(re.sub(r":(\w+)", r"(?P<\1>[\w-]+?)",
                      resource)) + r"(\?(?P<querystring>.*))?$")


class Route(object):
    """
    Route class
    """
    def __init__(self, resource):
        # syntax: /network/cellular/:id
        self.resource = trim_resource(resource)
        self.resource_regex = compile_resource(resource)
        self.handlers = []

        for method in ["get", "post", "put", "delete", "all"]:
            self.__setattr__(method, self.create_handler_func(method))

    def create_handler_func(self, method):
        """
        create_handler_func
        """
        def _handler(callback, schema=None):
            """
            _handler
            """
            # reentrant default is False [POST, DELETE, PUT]
            reentrant = False
            if method == "get":
                reentrant = True

            self.handlers.append({
                "method": method,
                "callback": callback,
                "schema": schema,
                "reentrant": reentrant
            })
            return self

        return _handler

    def dispatch(self, message):
        """
        dispatch
        """
        handlers = []
        for handler in self.handlers:
            if handler["method"] != message.method:
                continue

            handlers.append(handler)
        return handlers

    def get_methods(self):
        methods = set()
        for handler in self.handlers:
            methods.add(handler["method"])

        return list(methods)


class Router(object):
    """
    Router class
    """
    def __init__(self):
        self.routes = {}
        for method in ["get", "post", "put", "delete", "all"]:
            self.__setattr__(method, self.create_route_func(method=method))

    def route(self, resource):
        """
        route
        """
        route = self.routes.get(resource, Route(resource))
        self.routes.update({resource: route})
        return route

    def create_route_func(self, method):
        """
        create_route_func
        """
        def _route(resource, handler, schema=None):
            """
            _route
            """
            route = self.routes.get(resource, Route(resource))
            route.__getattribute__(method)(handler, schema)
            self.routes.update({resource: route})
            return self

        return _route

    def dispatch(self, message):
        """
        dispatch
        """
        results = []
        # match routes
        for resource, route in self.routes.items():
            __message = message.match(route)
            if __message is None:
                continue

            route_result = route.dispatch(__message)
            if len(route_result) == 0:
                continue

            results.append({
                "handlers": route_result,
                "message": __message
            })

        return results

    def get_routes(self):
        routes = {}
        for resource, route in self.routes.iteritems():
            routes.update({resource: route.get_methods()})

        return routes



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
This is s Sanji Object
"""

import inspect
import logging
import signal
import sys
import os
import threading
import re
import traceback
from random import random
from threading import Event
from threading import Thread
from time import sleep
from voluptuous import MultipleInvalid

from sanji.message import Message
from sanji.message import MessageType
from sanji.publish import Publish
from sanji.publish import Retry
from sanji.router import Router
from sanji.session import Session
from sanji.session import TimeoutError
from sanji.bundle import Bundle

try:
    from queue import Queue
except ImportError:
    from Queue import Queue

_logger = logging.getLogger("sanji.sdk")


class Sanji(object):
    """
    This is for sanji framework.
    """
    def __init__(self, *args, **kwargs):

        # Setup default options
        bundle = kwargs.get("bundle", None)
        connection = kwargs.get("connection", None)
        stop_event = kwargs.get("stop_event", Event())

        if connection is None:
            raise ValueError("Connection is required.")

        # Model-related
        bundle_dir = os.path.dirname(inspect.getfile(self.__class__))
        if bundle is None:
            bundle = Bundle(bundle_dir=bundle_dir)
        self.bundle = bundle
        self.stop_event = stop_event
        self.reg_thread = None
        self.reg_delay = lambda: random() * 5

        # Router-related (Dispatch)
        self.router = Router()
        self.dispatch_thread_count = 3
        self.thread_list = []

        # Response-related (Resolve)
        self._session = Session()
        self.resolve_thread_count = 1

        # Message Bus
        self._conn = connection
        self.is_ready = Event()
        self.req_queue = Queue()
        self.res_queue = Queue()

        # Setup callbacks
        self._conn.set_on_message(self.on_sanji_message)
        self._conn.set_on_connect(self.on_connect)
        self._conn.set_on_publish(self.on_publish)

        # Publisher
        self.publish = Publish(self._conn, self._session)

        # Register signal to call stop() (only mainthread could do this)
        if threading.current_thread().__class__.__name__ == '_MainThread':
            signal.signal(signal.SIGINT, self.exit)

        # Auto-register routes
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        self._register_routes(methods)

        # Custom init function
        if hasattr(self, 'init') and \
           hasattr(self.init, '__call__'):
            _logger.debug("Custom init start")
            self.init(*args, **kwargs)
            _logger.debug("Custom init finish")

    def _register_routes(self, methods):
        """
        _register_routes
        """
        # setup routes by decorator
        methods = [(n, v) for (n, v) in methods if v.__name__ == "wrapper"]
        methods = sorted(methods, key=lambda x: x[1]._order)
        for name, value in methods:
            value()  # execute setting route

        return methods

    def _dispatch_message(self):
        """
        _dispatch_message
        """
        while True:
            message = self.req_queue.get()
            if message is None:
                _logger.debug("_dispatch_message thread is terminated")
                return

            if message._type != MessageType.EVENT:
                self.__dispatch_message(message)
            elif message._type == MessageType.EVENT:
                self.__dispatch_event_message(message)

    def __dispatch_event_message(self, message):
        results = self.router.dispatch(message)

        def ___dispatch(handler, message):
            args_len = len(inspect.getargspec(handler["callback"]).args)
            if args_len == 2:
                handler["callback"](self, result["message"])

        try:
            for result in results:  # same route
                map(lambda handler: ___dispatch(handler, result["message"]),
                    result["handlers"])
        except Exception as e:
            _logger.error(e, exc_info=True)

    def __dispatch_message(self, message):
        results = self.router.dispatch(message)
        # Request not found
        if len(results) == 0:
            error_msg = "Route '%s' not found." % message.resource
            _logger.info(error_msg)
            _logger.debug(message.to_json())
            resp = self.publish.create_response(
                message, self.bundle.profile["name"])
            resp(code=404, data={"message": error_msg})
            return

        def ___dispatch(handler, message, resp):
            if handler["schema"] is not None:
                message.data = handler["schema"](message.data)
            args_len = len(inspect.getargspec(handler["callback"]).args)
            if args_len >= 3:
                handler["callback"](self, result["message"], resp)

        try:
            for result in results:  # same route
                resp = self.publish.create_response(
                    result["message"], self.bundle.profile["name"])
                map(lambda handler: ___dispatch(
                    handler, result["message"], resp
                    ),
                    result["handlers"])
        except Exception as e:
            _logger.error(e, exc_info=True)
            resp_data = {"message": "Internal Error."}

            if os.getenv("SANJI_ENV", 'debug') == 'debug':
                ex_type, ex, tb = sys.exc_info()
                resp_data["traceback"] = "".join(traceback.format_tb(tb))

            resp = self.publish.create_response(
                message, self.bundle.profile["name"])

            # if exception is belongs to schema error
            if isinstance(e, MultipleInvalid):
                return resp(code=400, data={"message": str(e)})

            return resp(code=500, data=resp_data)

    def _resolve_responses(self):
        """
        _resolve_responses
        """
        while True:
            message = self.res_queue.get()
            if message is None:
                _logger.debug("_resolve_responses thread is terminated")
                return
            self.__resolve_responses(message)

    def __resolve_responses(self, message):
        session = self._session.resolve(message.id, message)
        if session is None:
            self.req_queue.put(message.to_event())

    def on_publish(self, client, userdata, mid):
        with self._session.session_lock:
            self._session.resolve_send(mid)

    def _create_thread_pool(self):

        def stop(queue):
            def _stop():
                queue.put(None)
            return _stop

        # create a thread pool
        for _ in range(0, self.dispatch_thread_count):
            thread = Thread(target=self._dispatch_message,
                            name="thread-dispatch-%s" % _)
            thread.daemon = True
            thread.start()
            self.thread_list.append((thread, stop(self.req_queue)))

        for _ in range(0, self.resolve_thread_count):
            thread = Thread(target=self._resolve_responses,
                            name="thread-resolve-%s" % _)
            thread.daemon = True
            thread.start()
            self.thread_list.append((thread, stop(self.res_queue)))

        _logger.debug("Thread pool is created")

    def start(self):
        """
        start
        """
        def main_thread():
            # create resp, req thread pool
            self._create_thread_pool()
            # start connection, this will block until stop()
            self.conn_thread = Thread(target=self._conn.connect)
            self.conn_thread.daemon = True
            self.conn_thread.start()

            # register model to controller...
            self.is_ready.wait()

            if hasattr(self, 'run'):
                _logger.debug("Start running...")
                self.run()

        # start main_thread
        self.main_thread = Thread(target=main_thread)
        self.main_thread.daemon = True
        self.main_thread.start()

        if threading.current_thread().__class__.__name__ == '_MainThread':
            # control this bundle stop or not
            while not self.stop_event.wait(1):
                sleep(1)
        else:
            self.stop_event.wait()

        self.stop()
        _logger.debug("Shutdown successfully")

    def exit(self, signum=None, frame=None):
        """
        hook ctrl + c to exit program
        """
        self.stop()
        sys.exit(0)

    def stop(self, *args, **kwargs):
        """
        exit
        """
        _logger.debug("Bundle [%s] has been shutting down" %
                      self.bundle.profile["name"])

        if hasattr(self, 'before_stop') and \
           hasattr(self.before_stop, '__call__'):
            _logger.debug("Invoking before_stop...")
            self.before_stop()

        self._conn.disconnect()
        self._session.stop()
        self.stop_event.set()

        # TODO: shutdown all threads
        for thread, stop in self.thread_list:
            stop()
        for thread, stop in self.thread_list:
            thread.join()
        self.is_ready.clear()

    def on_sanji_message(self, client, userdata, msg):
        """This function will recevie all message from mqtt
        client
            the client instance for this callback
        userdata
            the private user data as set in Client() or userdata_set()
        message
            an instance of MQTTMessage. This is a class with members topic,
            payload, qos, retain.
        """
        try:
            message = Message(msg.payload)
        except (TypeError, ValueError) as e:
            _logger.error(e, exc_info=True)
            return

        if message.type() == MessageType.UNKNOWN:
            _logger.debug("Got an UNKNOWN message, don't dispatch")
            return

        if message.type() == MessageType.RESPONSE:
            self.res_queue.put(message)

        if message.type() == MessageType.REQUEST or \
           message.type() == MessageType.DIRECT or \
           message.type() == MessageType.HOOK or \
           message.type() == MessageType.EVENT:
            self.req_queue.put(message)

    def on_connect(self, client, userdata, flags, rc):
        """
        on_connect(self, client, obj, flags, rc):
        client
            the client instance for this callback
        userdata
            the private user data as set in Client() or userdata_set()
        flags
            response flags sent by the broker
        rc
            the connection result
        """
        _logger.debug("Connection established with result code %s" % rc)

        if self.reg_thread is not None and self.reg_thread.is_alive():
            _logger.debug("Joining previous reg_thread")
            self.reg_thread.join()

        def reg():
            delay = None
            if hasattr(self.reg_delay, '__call__'):
                delay = self.reg_delay()
            else:
                delay = self.reg_delay

            sleep(delay)
            self._conn.set_tunnels(self._conn.tunnels)
            model_profile = self.get_profile("model")
            view_profile = self.get_profile("view")
            self.deregister(model_profile)
            self.deregister(view_profile)
            self.register(model_profile)
            self.register(view_profile)
            self.is_ready.set()

        self.reg_thread = Thread(target=reg)
        self.reg_thread.daemon = True
        self.reg_thread.start()

    def register(self, reg_data, retry=True, interval=1, timeout=3):
        """
        register function
        retry
            True, infinity retries
            False, no retries
            Number, retries times
        interval
            time period for retry
        return
            False if no success
            Tunnel if success
        """
        if len(reg_data["resources"]) == 0:
            _logger.debug("%s no need to register due to no resources" %
                          (reg_data["name"]))
            return

        def _register():
            try:
                resp = self.publish.direct.post(
                    "/controller/registration", reg_data)
                if resp.code == 200:
                    return resp
            except TimeoutError:
                _logger.debug("Register message is timeout")

            return False

        resp = _register()
        while resp is False:
            _logger.debug("Register failed.")
            self.deregister(reg_data)
            resp = _register()

        if resp is None:
            _logger.error("Can\'t not register to controller")
            self.stop()
            return False

        self._conn.set_tunnel(
            reg_data["role"], resp.data["tunnel"], self.on_sanji_message)
        self.bundle.profile["currentTunnels"] = [
            tunnel for tunnel, callback in self._conn.tunnels.items()]
        self.bundle.profile["regCount"] = \
            self.bundle.profile.get("reg_count", 0) + 1

        _logger.debug("Register successfully %s tunnel: %s"
                      % (reg_data["name"], resp.data["tunnel"],))

    def deregister(self, reg_data, retry=True, interval=1, timeout=3):
        """
        Deregister model/view of this bundle
        """
        Retry(target=self.publish.direct.delete,
              args=("/controller/registration", reg_data,),
              kwargs={"timeout": timeout},
              options={"retry": retry, "interval": interval})
        _logger.debug("Deregister successfully %s tunnel: %s" %
                      (reg_data["name"],
                       self._conn.tunnels[reg_data["role"]][0],))

    def get_profile(self, role="model"):
        profile = dict((k, v) for k, v in self.bundle.profile.items())
        profile["tunnel"] = self._conn.tunnels["internel"][0]
        profile["resources"] = []
        profile["role"] = role
        profile["name"] = "%s%s" % \
            (profile["name"], '' if role == self.bundle.profile["role"]
                              else '-' + role,)

        for _ in self.bundle.profile["resources"]:
            if _.get("role", self.bundle.profile["role"]) != role:
                continue
            profile["resources"].append(re.sub(r":(\w+)", r"+", _["resource"]))

        return profile


def Route(resource=None, methods=["get", "post", "put", "delete"],
          schema=None):
    """
    route
    """
    def _route(func):
        def wrapper(self, *args, **kwargs):
            # "test" argument means no wrap func this time,
            # return original func immediately.
            if kwargs.get("test", False):
                kwargs.pop("test")
                func(self, *args, **kwargs)

            _methods = methods
            if isinstance(methods, str):
                _methods = [methods]
            route = self.router.route(resource)
            for method in _methods:
                getattr(route, method)(func, schema)
        # Ordered by declare sequence
        # http://stackoverflow.com/questions/4459531/how-to-read-class-attributes-in-the-same-order-as-declared
        f_locals = sys._getframe(1).f_locals
        _order = len([v for v in f_locals.itervalues()
                     if hasattr(v, '__call__') and
                     hasattr(v, '__name__') and
                     v.__name__ == "wrapper"])
        wrapper.__dict__["_order"] = _order
        return wrapper
    return _route
